/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.util.*;
import org.sablecc.sablecc.analysis.*;
import org.sablecc.sablecc.node.*;

public class ConstructNFA extends DepthFirstAdapter
{
  private ResolveIds ids;
  private String stateName;

  private Map<Node, Object> nodeValues; // values are of type CharSet or NFA
  private Map<PTokenDef, NFA> nfasByToken;
  private NFA fullNFA;

  ConstructNFA(ResolveIds ids, String stateName)
  {
    this.ids = ids;
    this.stateName = stateName;

    this.nodeValues = new HashMap<>();
    this.nfasByToken = new HashMap<>();
    this.fullNFA = null;
  }

  public NFA getNFA() {
    return this.fullNFA;
  }

  @Override
  public void outAHelperDef(AHelperDef node)
  {
    this.nodeValues.put(node, this.nodeValues.get(node.getRegExp()));

    // free memory
    this.nodeValues.remove(node.getRegExp());
  }

  @Override
  public void outATokens(ATokens node)
  {
    PTokenDef[] tokenDefs = node.getTokenDefs().toArray(new PTokenDef[0]);

    for(int i = tokenDefs.length - 1; i >= 0 ; i--)
    {
      NFA nfa = this.nfasByToken.get(tokenDefs[i]);
      if(nfa != null)
      {
        if(fullNFA == null)
        {
          fullNFA = nfa;
        }
        else
        {
          fullNFA = nfa.merge(fullNFA);
        }

        // free memory
        this.nfasByToken.remove(tokenDefs[i]);
      }
    }
  }

  @Override
  public void outATokenDef(ATokenDef node)
  {
    Set<String> set = stateNamesFromStateList(node.getStateList());
    Object o1 = this.nodeValues.get(node.getRegExp());

    if(set.isEmpty() || set.contains(stateName))
    {
      //System.out.print("*");

      NFA n1 = (o1 instanceof NFA) ? (NFA) o1 : new NFA((CharSet) o1);
      String name = ids.names.get(node);

      n1.states[n1.states.length - 1].accept = name;
      this.nfasByToken.put(node, n1);
    }
    else
    {
      //System.out.print("-");
    }

    // free memory
    this.nodeValues.remove(node.getRegExp());
  }

  private static Set<String> stateNamesFromStateList(PStateList stateList)
  {
    if (stateList == null) {
      return Collections.emptySet();
    }

    final AStateList node = (AStateList)stateList;
    Set<String> set = new TreeSet<>();
    AStateListTail[] stateListTails = node.getStateLists().toArray(new AStateListTail[0]);

    for(int i = stateListTails.length - 1; i >= 0 ; i--)
    {
      String str = stateListTails[i].getId().getText().toUpperCase();
      set.add(str);
    }

    set.add(node.getId().getText().toUpperCase());
    return set;
  }

  @Override
  public void outARegExp(ARegExp node)
  {
    PConcat[] concats = node.getConcats().toArray(new PConcat[0]);
    NFA result = null;

    if(concats.length > 1)
    {
      for(int i = concats.length - 1; i >= 0 ; i--)
      {
        Object o = this.nodeValues.get(concats[i]);
        NFA nfa = (o instanceof NFA) ? (NFA) o : new NFA((CharSet) o);

        if(result == null)
        {
          result = nfa;
        }
        else
        {
          result = nfa.alternate(result);
        }

        // free memory
        this.nodeValues.remove(concats[i]);
      }
      this.nodeValues.put(node, result);
    }
    else if(concats.length == 1)
    {
      this.nodeValues.put(node, this.nodeValues.get(concats[0]));

      // free memory
      this.nodeValues.remove(concats[0]);
    }
  }

  @Override
  public void outAConcat(AConcat node)
  {
    PUnExp[] unExps = node.getUnExps().toArray(new PUnExp[0]);

    if(unExps.length == 0)
    {
      this.nodeValues.put(node, new NFA());
    }
    else if(unExps.length == 1)
    {
      this.nodeValues.put(node, this.nodeValues.get(unExps[0]));

      // free memory
      this.nodeValues.remove(unExps[0]);
    }
    else
    {
      NFA result = null;

      for(int i = unExps.length - 1; i >= 0 ; i--)
      {
        Object o = this.nodeValues.get(unExps[i]);
        NFA nfa = (o instanceof NFA) ? (NFA) o : new NFA((CharSet) o);

        if(result == null)
        {
          result = nfa;
        }
        else
        {
          result = nfa.concatenate(result);
        }

        // free memory
        this.nodeValues.remove(unExps[i]);
      }

      this.nodeValues.put(node, result);
    }
  }

  @Override
  public void outAUnExp(AUnExp node)
  {
    Object o = this.nodeValues.get(node.getBasic());

    if(node.getUnOp() instanceof AStarUnOp)
    {
      NFA n = (o instanceof NFA) ? (NFA) o : new NFA((CharSet) o);
      this.nodeValues.put(node, n.zeroOrMore());
    }
    else if(node.getUnOp() instanceof AQMarkUnOp)
    {
      NFA n = (o instanceof NFA) ? (NFA) o : new NFA((CharSet) o);
      this.nodeValues.put(node, n.zeroOrOne());
    }
    else if(node.getUnOp() instanceof APlusUnOp)
    {
      NFA n = (o instanceof NFA) ? (NFA) o : new NFA((CharSet) o);
      this.nodeValues.put(node, n.oneOrMore());
    }
    else
    {
      this.nodeValues.put(node, o);
    }

    // free memory
    this.nodeValues.remove(node.getBasic());
  }

  @Override
  public void outACharBasic(ACharBasic node)
  {
    char c = charValue(node.getChar());
    this.nodeValues.put(node, new CharSet(c));
  }

  @Override
  public void outASetBasic(ASetBasic node)
  {
    this.nodeValues.put(node, this.nodeValues.get(node.getSet()));

    // free memory
    this.nodeValues.remove(node.getSet());
  }

  @Override
  public void outAStringBasic(AStringBasic node)
  {
    String s = node.getString().getText();
    s = s.substring(1, s.length() -1);

    this.nodeValues.put(node, new NFA(s));
  }

  @Override
  public void outAIdBasic(AIdBasic node)
  {
    Object o = this.nodeValues.get(ids.helpers.get(node.getId().getText()));

    if(o instanceof NFA)
    {
      this.nodeValues.put(node, ((NFA) o).clone());
    }
    else
    {
      this.nodeValues.put(node, ((CharSet) o).clone());
    }
  }

  @Override
  public void outARegExpBasic(ARegExpBasic node)
  {
    this.nodeValues.put(node, this.nodeValues.get(node.getRegExp()));

    // free memory
    this.nodeValues.remove(node.getRegExp());
  }

  private static char charValue(final PChar node) {
    if(node instanceof ACharChar)
    {
      return ((ACharChar)node).getChar().getText().charAt(1);
    }
    else if(node instanceof ADecChar)
    {
      return (char)Integer.parseInt(((ADecChar)node).getDecChar().getText());
    }
    else if(node instanceof AHexChar)
    {
      return (char)Integer.parseInt(((AHexChar)node).getHexChar().getText().substring(2), 16);
    }
    else
    {
      throw new IllegalArgumentException("Unhandled char type: " + node.getClass());
    }
  }

  @Override
  public void outAOperationSet(AOperationSet node)
  {
    try
    {
      CharSet cs1 = (CharSet) this.nodeValues.get(node.getLeft());
      CharSet cs2 = (CharSet) this.nodeValues.get(node.getRight());

      if(node.getBinOp() instanceof APlusBinOp)
      {
        this.nodeValues.put(node, cs1.union(cs2));
      }
      else if(node.getBinOp() instanceof AMinusBinOp)
      {
        this.nodeValues.put(node, cs1.diff(cs2));
      }
    }
    catch(RuntimeException e)
    {
      throw new RuntimeException(node + " is invalid.", e);
    }

    // free memory
    this.nodeValues.remove(node.getLeft());
    this.nodeValues.remove(node.getRight());
  }

  @Override
  public void outAIntervalSet(AIntervalSet node)
  {
    char c1 = charValue(node.getLeft());
    char c2 = charValue(node.getRight());

    if(c1 > c2)
    {
      throw new RuntimeException(node + " is invalid.");
    }

    this.nodeValues.put(node, new CharSet(c1, c2));
  }
}

